// Password.cs
//
// Class implementation for password management
//

using System;
using System.Diagnostics;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Runtime.InteropServices;


namespace Rsdn
{
	/// <summary>
    /// The class that encapsulates user password
	/// </summary>
    [ClassInterface(ClassInterfaceType.AutoDual)]
    public sealed class Password
	{
		/// <summary>
		/// Inits object using password correction and password hash
		/// </summary>
		/// <param name="salt">Password correction</param>
		/// <param name="hash">Password hash</param>
		public Password(byte[] salt, byte[] hash)
		{
			_salt = (byte[])salt.Clone();
			_hash = (byte[])hash.Clone();
		}

		/// <summary>
        /// Inits object using password correction and password hash
		/// </summary>
        /// <param name="salt">Base64-encoded password correction</param>
		/// <param name="hash">Base64-encoded password hash</param>
		public Password(string salt, string hash)
		{
			_salt = Convert.FromBase64String(salt);
			_hash = Convert.FromBase64String(hash);
		}

		/// <summary>
		/// Inits object using given plain text password
		/// </summary>
		/// <param name="clearText">Plain text password (not crypted)</param>
        /// <remarks>For security reasons calling process should zeroize this array as soon as possible</remarks>
		public Password(char[] clearText)
		{
			_salt = GenerateRandom(6);
			_hash = HashPassword(clearText);
		}

        /// <summary>
        /// Empty constructor, generates random password (but we'll never know what ;)
        /// </summary>
        public Password()
        {
            _salt = GenerateRandom(6);
            _hash = HashPassword(Generate());
        }

        /// <summary>
        /// Reinitializes password class object using plain password as an argument
        /// </summary>
        /// <param name="clearText">Palin text password</param>
        public void ResetPassword(char[] clearText)
        {
            _salt = GenerateRandom(6);
            _hash = HashPassword(clearText);
        }
        
        /// <summary>
		/// Returns password correction as a base64-encoded string
		/// </summary>
		public string Salt
		{
			get { return Convert.ToBase64String(_salt); }
		}

		/// <summary>
        /// Returns password correction as a byte array
		/// </summary>
		public byte[] RawSalt
		{
			get { return (byte[])_salt.Clone(); }
		}

		/// <summary>
        /// Returns password hash as a base64-encoded string
		/// </summary>
		public string Hash
		{
			get { return Convert.ToBase64String(_hash); }
		}

		/// <summary>
        /// Returns password hash as a byte array
		/// </summary>
		public byte[] RawHash
		{
			get { return (byte[])_hash.Clone(); }
		}

		/// <summary>
		/// Checks whether given plain text password matches to password correction and hash stored in the object 
		/// </summary>
		/// <param name="clearText">Plain text password</param>
		/// <returns>Method returns true if given password match object password correction and hash, and false - otherwise</returns>
		public bool	Verify(char[] clearText)
		{
			byte[] hash = HashPassword(clearText);
			
			if (hash.Length	== _hash.Length)
			{
				for (int i = 0; i < hash.Length; i++)
				{
					if (hash[i] != _hash[i])
						return false;
				}

				return true;
			}

			return false;
		}

		/// <summary>Generates random password</summary>
        /// <returns>Generated password as char array</returns>
		public static char[] Generate()
		{
			char[] random = new char[12];
            // generate 9 random bytes; this is enough to get 12 random symbols from base64 set
			byte[] rnd = GenerateRandom(9);

            // convert random bytes to base64
			Convert.ToBase64CharArray(rnd, 0, rnd.Length, random, 0);

            // clean up  working array
			Array.Clear(rnd, 0, rnd.Length);

			return random;
		}

		/// <summary>
		/// Creates random bytes array with given length
		/// </summary>
		/// <param name="size">Array length</param>
        /// <returns>Random bytes array </returns>
		private static byte[] GenerateRandom(int size)
		{
			byte[] random = new byte[size];
			RandomNumberGenerator.Create().GetBytes(random);
			return random;
		}

		/// <summary>
		/// Hashes given plain text password using password correction (_salt)
		/// </summary>
		/// <param name="clearText">Plain text password</param>
		/// <returns>Password hash</returns>
		private byte[] HashPassword(char[] clearText)
		{
			Encoding utf8 = Encoding.UTF8;
			byte[] hash;

            // create enough length working array
			byte[] data = new byte[_salt.Length + utf8.GetMaxByteCount(clearText.Length)];

			try
			{
                // copy salt to working array
				Array.Copy(_salt, 0, data, 0, _salt.Length);

				// copy password to working array convrting it to UTF-8
				int byteCount = utf8.GetBytes(clearText, 0, clearText.Length,
					data, _salt.Length);

                // hash array data
				using (HashAlgorithm alg = new SHA256Managed())
					hash = alg.ComputeHash(data, 0, _salt.Length + byteCount);
			}
			finally
			{
                // clean up working array in the end of work to avoid password leak
				Array.Clear(data, 0, data.Length);
			}

			return hash;
		}

		#region	Private	data members

        private byte[] _salt;		// salt (random password correction)
		private	byte[] _hash;		// password hash

		#endregion

        #region COM interop section

        /// <summary>
        /// COM interop registering function
        /// </summary>
        /// <param name="t"></param>
        [ComRegisterFunctionAttribute]
        public static void RegisterFunction(Type t)
        {
            Microsoft.Win32.Registry.ClassesRoot.CreateSubKey("CLSID\\{" + t.GUID.ToString().ToUpper() + "}\\Programmable");
        }

        /// <summary>
        /// COM interop unregistering function
        /// </summary>
        /// <param name="t"></param>
        [ComUnregisterFunctionAttribute]
        public static void UnregisterFunction(Type t)
        {
            Microsoft.Win32.Registry.ClassesRoot.DeleteSubKey("CLSID\\{" + t.GUID.ToString().ToUpper() + "}\\Programmable");
        }
        
        #endregion
    }
}
